from __future__ import absolute_import, division, print_function

import sys
import re
import os
import gzip
import bz2
import uuid
import csv
from tempfile import NamedTemporaryFile

from glob import glob
from contextlib import contextmanager

from toolz import concat, keyfilter, keymap, merge, valfilter

import pandas as pd

from dask.threaded import get as dsk_get
import datashape

from datashape import discover, Record, Option
from datashape.predicates import isrecord
from datashape.dispatch import dispatch

from ..compatibility import unicode, PY2
from ..utils import keywords, ext, sample, tmpfile
from ..append import append
from ..convert import convert, ooc_types
from ..resource import resource
from ..chunks import chunks
from ..temp import Temp
from ..numpy_dtype import dshape_to_pandas
from .pandas import coerce_datetimes
from functools import partial

dialect_terms = '''delimiter doublequote escapechar lineterminator quotechar
quoting skipinitialspace strict'''.split()

aliases = {'sep': 'delimiter'}


class PipeSniffer(csv.Sniffer):
    """A csv sniffer that adds the ``'|'`` character to the list of preferred
    delimiters and gives space the least precedence.

    Notes
    -----
    The :module:`~csv` module has a list of preferred delimiters that will be
    returned before the delimiter(s) discovered by parsing the file. This
    results in a dialect having an incorrect delimiter if the data contain a
    non preferred delimiter and a preferred one. See the Examples section for
    an example showing the difference between the two sniffers.

    Examples
    --------
    >>> import csv
    >>> data = 'a|b| c|d'  # space is preferred but '|' is more common
    >>> csv.Sniffer().sniff(data).delimiter  # incorrect
    ' '
    >>> PipeSniffer().sniff(data).delimiter  # correct
    '|'
    >>> data = 'a|b|Nov 10, 1981|d'
    >>> csv.Sniffer().sniff(data).delimiter  # can't handle every case :(
    ','
    """
    def __init__(self, *args, **kwargs):
        csv.Sniffer.__init__(self, *args, **kwargs)
        self.preferred = [',', '\t', ';', '|', ':', ' ']


def alias(key):
    """ Alias kwarg dialect keys to normalized set

    >>> alias('sep')
    'delimiter'
    """
    key = key.lower()
    return aliases.get(key, key)


class ModeProxy(object):
    """Gzipped files use int enums for mode so we want to proxy it to provide
    The proper string version.
    """
    def __init__(self, file, mode):
        self._file = file
        self.mode = mode

    def __getattr__(self, attr):
        return getattr(self._file, attr)

    def __iter__(self):
        return iter(self._file)


@contextmanager
def open_file(path, *args, **kwargs):
    f = compressed_open.get(ext(path), open)(path, *args, **kwargs)
    try:
        yield ModeProxy(f, kwargs.get('mode', 'r'))
    finally:
        f.close()


def infer_header(path, nbytes=10000, encoding='utf-8', **kwargs):
    if encoding is None:
        encoding = 'utf-8'
    with open_file(path, 'rb') as f:
        raw = f.read(nbytes)
    if not raw:
        return True
    sniffer = PipeSniffer()
    decoded = raw.decode(encoding, 'replace')
    sniffable = decoded.encode(encoding) if PY2 else decoded
    try:
        return sniffer.has_header(sniffable)
    except csv.Error:
        return None


def newlines(encoding):
    return b'\r\n'.decode(encoding), b'\n'.decode(encoding)


def sniff_dialect(path, nbytes, encoding='utf-8'):
    if not os.path.exists(path):
        return {}
    if encoding is None:
        encoding = 'utf-8'
    with open_file(path, 'rb') as f:
        raw = f.read(nbytes).decode(encoding or 'utf-8', 'replace')
    sniffer = PipeSniffer()
    try:
        dialect = sniffer.sniff(raw, delimiters=sniffer.preferred)
    except csv.Error:
        return {}
    crnl, nl = newlines(encoding)
    dialect.lineterminator = crnl if crnl in raw else nl
    return dialect_to_dict(dialect)


def dialect_to_dict(dialect):
    return dict((name, getattr(dialect, name))
                for name in dialect_terms if hasattr(dialect, name))


class NoCloseFile(object):
    def __init__(self, buffer):
        self._buf = buffer

    def __getattr__(self, attr):
        return getattr(self._buf, attr)

    def close(self):
        pass

    def __enter__(self):
        return self._buf

    def __exit__(self, *exc_info):
        pass

    def __iter__(self):
        return iter(self._buf)


class CSV(object):

    """ Proxy for a CSV file

    Parameters
    ----------

    path : str
        Path to file on disk
    has_header : bool
        If the csv file has a header or not
    encoding : str (default utf-8)
        File encoding
    kwargs : other...
        Various choices about dialect
    """
    canonical_extension = 'csv'

    def __init__(self, path, has_header=None, encoding='utf-8',
                 sniff_nbytes=10000, buffer=None, **kwargs):
        self.path = path
        self._has_header = has_header
        self.encoding = encoding or 'utf-8'
        self._kwargs = kwargs
        self._sniff_nbytes = sniff_nbytes
        self._buffer = buffer

        if path is not None and buffer is not None:
            raise ValueError("may only pass one of 'path' or 'buffer'")

        if path is None and buffer is None:
            raise ValueError("must pass one of 'path' or 'buffer'")

    def _sniff_dialect(self, path):
        kwargs = self._kwargs
        dialect = sniff_dialect(path, self._sniff_nbytes,
                                encoding=self.encoding)
        kwargs = merge(dialect, keymap(alias, kwargs))
        return valfilter(lambda x: x is not None,
                         dict((d, kwargs[d])
                              for d in dialect_terms if d in kwargs))

    @property
    def dialect(self):
        with sample(self) as fn:
            return self._sniff_dialect(fn)

    @property
    def has_header(self):
        if self._has_header is None:
            with sample(self) as fn:
                with open(fn, mode='rb') as f:
                    raw = f.read()
                self._has_header = not raw or infer_header(fn,
                                                           self._sniff_nbytes,
                                                           encoding=self.encoding)
        return self._has_header

    def open(self, mode='rb', **kwargs):
        buf = self._buffer
        if buf is not None:
            return NoCloseFile(buf)

        return open_file(self.path, mode=mode, **kwargs)


@sample.register(CSV)
@contextmanager
def sample_csv(csv, length=8192, **kwargs):
    if csv.path is None or not os.path.exists(csv.path):
        mgr = NamedTemporaryFile(suffix='.csv', mode='wb+')
    else:
        mgr = csv.open(mode='rb')

    with mgr as f:
        with tmpfile(csv.canonical_extension) as fn:
            with open(fn, mode='wb') as tmpf:
                tmpf.write(f.read(length))
            yield fn


@append.register(CSV, object)
def append_object_to_csv(c, seq, **kwargs):
    append(c, convert(chunks(pd.DataFrame), seq, **kwargs), **kwargs)
    return c


compressed_open = {'gz': gzip.open, 'bz2': bz2.BZ2File}


@append.register(CSV, pd.DataFrame)
def append_dataframe_to_csv(c, df, dshape=None, **kwargs):
    if not os.path.exists(c.path) or not os.path.getsize(c.path):
        has_header = kwargs.pop('header', c.has_header)
    else:
        has_header = False
    sep = kwargs.get('sep',
                     kwargs.get('delimiter', c.dialect.get('delimiter', ',')))
    encoding = kwargs.get('encoding', c.encoding)

    if ext(c.path) in compressed_open:
        if sys.version_info[0] >= 3:
            kwargs['mode'] = 'at'
            kwargs['encoding'] = encoding
        elif sys.version_info[0] == 2:
            kwargs['mode'] = 'ab' if sys.platform == 'win32' else 'at'

    with c.open(mode=kwargs.get('mode', 'a')) as f:
        df.to_csv(f,
                  header=has_header,
                  index=False,
                  sep=sep,
                  encoding=encoding)

    return c


@append.register(CSV, chunks(pd.DataFrame))
def append_iterator_to_csv(c, cs, **kwargs):
    for chunk in cs:
        append(c, chunk, **kwargs)
    return c


@convert.register(pd.DataFrame, (Temp(CSV), CSV), cost=20.0)
def csv_to_dataframe(c, dshape=None, chunksize=None, nrows=None, **kwargs):
    try:
        return _csv_to_dataframe(c, dshape=dshape, chunksize=chunksize,
                                 nrows=nrows, **kwargs)
    except StopIteration:
        if nrows:
            return _csv_to_dataframe(c, dshape=dshape, chunksize=chunksize,
                                     **kwargs)
        else:
            raise


def _csv_to_dataframe(c, dshape=None, chunksize=None, **kwargs):
    header = {False: None, True: 0}.get(
        kwargs.pop('has_header', c.has_header), 'infer')

    sep = kwargs.pop(
        'sep', kwargs.pop('delimiter', c.dialect.get('delimiter', ',')))
    encoding = kwargs.pop('encoding', c.encoding)

    if dshape:
        dtypes, parse_dates = dshape_to_pandas(dshape)
        if isrecord(dshape.measure):
            names = kwargs.get('names', dshape.measure.names)
        else:
            names = kwargs.get('names')
    else:
        dtypes = parse_dates = names = None

    usecols = kwargs.pop('usecols', None)
    if parse_dates and usecols:
        parse_dates = [col for col in parse_dates if col in usecols]

    # See read_csv docs for header for reasoning
    if names:
        try:
            with c.open() as f:
                found_names = pd.read_csv(f,
                                          nrows=1,
                                          encoding=encoding,
                                          sep=sep)
        except StopIteration:
            with c.open() as f:
                found_names = pd.read_csv(f, encoding=encoding, sep=sep)
    if names and header == 'infer':
        if [n.strip() for n in found_names] == [n.strip() for n in names]:
            header = 0
        elif (all(re.match('^\s*\D\w*\s*$', n) for n in found_names) and
                not all(dt == datashape.string for dt in dshape.measure.types)):
            header = 0
        else:
            header = None

    kwargs = keyfilter(keywords(pd.read_csv).__contains__, kwargs)
    with c.open() as f:
        return pd.read_csv(f,
                           header=header,
                           sep=sep,
                           encoding=encoding,
                           dtype=dtypes,
                           parse_dates=parse_dates,
                           names=names,
                           chunksize=chunksize,
                           usecols=usecols,
                           **kwargs)


@convert.register(chunks(pd.DataFrame), (Temp(CSV), CSV), cost=10.0)
def CSV_to_chunks_of_dataframes(c, chunksize=2 ** 20, **kwargs):
    # Load a small 1000 line DF to start
    # This helps with rapid viewing of a large CSV file
    first = csv_to_dataframe(c, nrows=1000, **kwargs)
    if len(first) == 1000:
        rest = csv_to_dataframe(
            c, chunksize=chunksize, skiprows=1000, **kwargs)
    else:
        rest = []

    data = [first] + rest
    return chunks(pd.DataFrame)(data)


@discover.register(CSV)
def discover_csv(c, nrows=1000, **kwargs):
    df = csv_to_dataframe(c, nrows=nrows, **kwargs)
    df = coerce_datetimes(df)

    columns = [str(c) if not isinstance(c, (str, unicode)) else c
               for c in df.columns]
    df.columns = [c.strip() for c in columns]

    # Replace np.nan with None. Forces type string rather than float
    for col in df.columns:
        if not df[col].count():
            df[col] = None

    measure = discover(df).measure

    # Use Series.notnull to determine Option-ness
    measure = Record([[name, Option(typ)
                       if df[name].isnull().any() and
                       not isinstance(typ, Option) else typ]
                      for name, typ in zip(measure.names, measure.types)])

    return datashape.var * measure


@resource.register('.+\.(csv|tsv|ssv|data|dat)(\.gz|\.bz2?)?')
def resource_csv(uri, **kwargs):
    return CSV(uri, **kwargs)


@resource.register('.*\*.+', priority=12)
def resource_glob(uri, **kwargs):
    filenames = sorted(glob(uri))
    r = resource(filenames[0], **kwargs)
    return chunks(type(r))([resource(u, **kwargs) for u in sorted(glob(uri))])


@convert.register(chunks(pd.DataFrame), (chunks(CSV), chunks(Temp(CSV))),
                  cost=10.0)
def convert_glob_of_csvs_to_chunks_of_dataframes(csvs, **kwargs):
    f = partial(convert, chunks(pd.DataFrame), **kwargs)

    def df_gen():
        # build up a dask graph to run all of the `convert` calls concurrently

        # use a list to hold the requested key names to ensure that we return
        # the results in the correct order
        p = []
        dsk = {}
        for n, csv_ in enumerate(csvs):
            key = 'p%d' % n
            dsk[key] = f, csv_
            p.append(key)

        return concat(dsk_get(dsk, p))

    return chunks(pd.DataFrame)(df_gen)


@convert.register(Temp(CSV), (pd.DataFrame, chunks(pd.DataFrame)))
def convert_dataframes_to_temporary_csv(data, **kwargs):
    fn = '.%s.csv' % uuid.uuid1()
    csv = Temp(CSV)(fn, **kwargs)
    return append(csv, data, **kwargs)


@dispatch(CSV)
def drop(c):
    if c.path is not None:
        os.unlink(c.path)


ooc_types.add(CSV)



from __future__ import absolute_import, division, print_function

try:
    import h5py  #  h5py has precedence over pytables
except:
    pass
from multipledispatch import halt_ordering, restart_ordering

halt_ordering()  # Turn off multipledispatch ordering

from .utils import ignoring
from .convert import convert
from .append import append
from .resource import resource
from .directory import Directory
from .into import into
from .odo import odo
from .create import create
from .drop import drop
from .temp import Temp
from .backends.text import TextFile
from .chunks import chunks, Chunks
from datashape import discover, dshape
import numpy as np

with ignoring(ImportError):
    from .backends.sas import sas7bdat
with ignoring(ImportError):
    from .backends.pandas import pd
with ignoring(ImportError):
    from .backends.bcolz import bcolz
with ignoring(ImportError):
    from .backends.h5py import h5py
with ignoring(ImportError):
    from .backends.hdfstore import HDFStore
with ignoring(ImportError):
    from .backends.pytables import PyTables
with ignoring(ImportError):
    from .backends import sql
with ignoring(ImportError):
    from .backends import mongo
with ignoring(ImportError):
    from .backends.csv import CSV
with ignoring(ImportError):
    from .backends.json import JSON, JSONLines
with ignoring(ImportError):
    from .backends.hdfs import HDFS
with ignoring(ImportError):
    from .backends.ssh import SSH
with ignoring(ImportError):
    from .backends import sql_csv
with ignoring(ImportError):
    from .backends.aws import S3
with ignoring(ImportError):
    from .backends import sql_csv
with ignoring(ImportError):
    from .backends.bokeh import ColumnDataSource
with ignoring(ImportError):
    from .backends.spark import RDD
with ignoring(ImportError):
    from .backends.sparksql import SparkDataFrame
with ignoring(ImportError):
    from .backends.url import URL
with ignoring(ImportError):
    from .backends.dask import dask


restart_ordering()  # Restart multipledispatch ordering and do ordering


from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



from __future__ import absolute_import, division, print_function

import tempfile
from contextlib import contextmanager
import shutil
from odo.backends.csv import CSV
from odo.directory import Directory, discover, resource, _Directory
from odo import into
from datashape import dshape
import os

@contextmanager
def csvs(n=3):
    path = tempfile.mktemp()
    os.mkdir(path)

    fns = [os.path.join(path, 'file_%d.csv' % i) for i in range(n)]

    for i, fn in enumerate(fns):
        into(fn, [{'a': i, 'b': j} for j in range(5)])

    try:
        yield path + os.path.sep
    finally:
        shutil.rmtree(path)


def test_discover():
    with csvs() as path:
        d = Directory(CSV)(path)
        assert discover(d) == dshape('var * {a: int64, b: int64}')


def test_resource_directory():
    with csvs() as path:
        r = resource(path)
        assert type(r) == Directory(CSV)
        assert r.path.rstrip(os.path.sep) == path.rstrip(os.path.sep)

        r2 = resource(os.path.join(path, '*.csv'))
        assert type(r2) == Directory(CSV)
        assert r2.path.rstrip(os.path.sep) == path.rstrip(os.path.sep)


def test_resource_directory():
    assert isinstance(resource(os.path.join('a', 'nonexistent', 'directory') +
                               os.path.sep),
                      _Directory)



from __future__ import absolute_import, division, print_function

import pytest

import numpy as np

from odo import odo, into
from odo.utils import tmpfile, filetext
from odo.backends.csv import CSV


def test_into_convert():
    assert odo((1, 2, 3), list) == [1, 2, 3]


def test_into_append():
    lst = []
    result = odo((1, 2, 3), lst)
    assert result == [1, 2, 3]
    assert result is lst


def test_into_append_failure():
    nd = np.array([1, 2, 3])
    with pytest.raises(TypeError):
        odo((4, 5), nd)


def test_into_curry():
    assert callable(into(list))
    data = (1, 2, 3)
    assert into(list)(data) == odo(data, list)


@pytest.mark.parametrize('f', [lambda x: u'%s' % x, lambda x: x])
def test_into_double_string(f):
    with filetext('alice,1\nbob,2', extension='.csv') as source:
        assert odo(source, list) == [('alice', 1), ('bob', 2)]

        with tmpfile('.csv') as target:
            csv = odo(source, f(target))
            assert isinstance(csv, CSV)
            with open(target, 'rU') as f:
                assert 'alice' in f.read()


@pytest.mark.parametrize('f', [lambda x: u'%s' % x, lambda x: x])
def test_into_string_on_right(f):
    with filetext('alice,1\nbob,2', extension='.csv') as source:
        assert odo(f(source), []) == [('alice', 1), ('bob', 2)]


def test_into_string_dshape():
    np.testing.assert_array_equal(odo([1, 2, 3], np.ndarray,
                                      dshape='var * float64'),
                                  np.array([1, 2, 3], dtype='float64'))


@pytest.mark.parametrize('dshape', [1, object()])
def test_into_invalid_dshape(dshape):
    with pytest.raises(TypeError):
        into(list, (1, 2, 3), dshape=dshape)



from __future__ import absolute_import, division, print_function

import os
from odo.resource import resource
from odo.utils import raises


def test_raises_not_implemented_error():
    assert raises(NotImplementedError,
            lambda: resource('5sdjkg9yg35420shfg083.3923.925y2560!'))



from __future__ import absolute_import, division, print_function

from odo.temp import Temp, _Temp
from odo.drop import drop

class Foo(object):
    pass


flag = [0]


@drop.register(Foo)
def drop_foo(foo, **kwrags):
    flag[0] = flag[0] + 1


def test_Temp():
    foo = Temp(Foo)()

    assert isinstance(foo, Foo)
    assert isinstance(foo, _Temp)

    x = flag[0]
    del foo
    assert flag[0] == x + 1
